import log
logger = log.get(__name__)

class Processor:
    
    def __init__(self, registry):
        self._registry = registry
        self._messageHandlers = {
            "policies": self._acceptPolicies,
            "endpoints": self._acceptEndpoints,
            "peers": self._acceptPeers
        }
        self._peerHandler = {
            "service": self._handleServicePeers,
            "cluster": self._handleServicePeers,
            "database": self._handleNativePeers,
            "queue": self._handleNativePeers
        }

    def accept(self, section, data):
        logger.debug('Accept Section: %s, data: %s', section, data)

        handler = self._messageHandlers[section]
        if handler is not None: 
            return handler(data)

    def _acceptPolicies(self, data):
        if data is not None:
            self._registry.set('policies', [], data)
        else:
            self._registry.reset('policies')

    def _acceptEndpoints(self, data):
        if data is not None:
            self._registry.set('endpoints', [], data)
            for endpointName, endpoints in data.items():
                self._registry.set('endpoints', [endpointName], endpoints)
        else:
            self._registry.reset('endpoints')

    def _acceptPeers(self, data):
        if data is not None:
            self._registry.set('endpoints', [], data)
            for kind, peers in data.items():
                if kind in self._peerHandler:
                    self._peerHandler[kind](kind, peers)
        else:
            self._registry.reset('service')
            self._registry.reset('cluster')

    def _handleServicePeers(self, kind, data):
        for name, serviceData in data.items():
            for endpoint, endpointData in serviceData.items():
                self._registry.set(kind, [name, endpoint], endpointData)

    def _handleNativePeers(self, kind, data):
        for name, endpointData in data.items():
            self._registry.set(kind, [name], endpointData)



import log
logger = log.get(__name__)

import starter as _starter
from functools import wraps
import inspect

# PEERS 
def monitorPeers(kind, name, endpoint, cb):
    _starter.registry.subscribe(kind, [name, endpoint], cb)

def getPeers(kind, name, endpoint):
    return _starter.registry.get(kind, [name, endpoint])

def getRandomPeer(kind, name, endpoint):
    peers = getPeers(kind, name, endpoint)
    return _starter.randomFromDict(peers)

def request(kind, name, endpoint):
    return _starter.makeRequest(kind, name, endpoint)

# DATABASES 
def monitorDatabases(name, cb):
    _starter.monitorNatives('database', name, cb)

def getDatabases(name):
    return _starter.getNatives('database', name)

def getDatabase(name):
    return _starter.getNative('database', name)

def getDatabaseClient(name):
    return _starter.getNativeClient('database', name)

# QUEUES 
def monitorQueues(name, cb):
    _starter.monitorNatives('queue', name, cb)

def getQueues(name):
    return _starter.getNatives('queue', name)

def getQueue(name):
    return _starter.getNative('queue', name)

def getQueueClient(name):
    return _starter.getNativeClient('queue', name)

# TRACING
def instrument(method, binary_annotations=None):
    return _starter.instrument(method, binary_annotations)

def instrument_func(**instArgs):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            mod = inspect.getmodule(f)
            binary_annotations = instArgs.get('binary_annotations', {})
            binary_annotations['source.path'] = mod.__file__
            binary_annotations['source.func'] = f.__name__
            binary_annotations['source.module'] = mod.__name__
            method = instArgs.get('name')
            if not method:
                method = 'func-' + f.__name__

            with instrument(method, binary_annotations):
                return f(*args, **kwargs)
        return wrapper
    return decorator

# FRAMEWORKS
def setupFlask(app):
    return _starter.setupFlask(app)


import log
logger = log.get(__name__)

logger.info('Starting...')

import sys
import json
import random as rand

from registry import Registry
from policy import Policy
from zipkin import Zipkin
from processor import Processor
from client import Client
from executor import Executor

import copy
import requests

registry = Registry()
policy = Policy(registry)
zipkin = Zipkin(policy)
processor = Processor(registry)

def onMessage(msg):
    for section, data in msg.items():
        # logger.info('Section %s, data: %s', section, data)
        processor.accept(section, data)
    # logger.info('**** REGISTRY: %s', json.dumps(registry.extractRoot(), indent=4, sort_keys=True))
client = Client(onMessage)


def monitorNatives(kind, name, cb):
    registry.subscribe(kind, [name], cb)

def getNatives(kind, name):
    return registry.get(kind, [name])

def getNative(kind, name):
    peers = getNatives(kind, name)
    return randomFromDict(peers)



def makeRequest(kind, name, endpoint):
    return RequestWrapper([kind, name, endpoint])

class RequestWrapper(object):

    def __init__(self, target):
        object.__setattr__(self, "_target", target)

    def __getattribute__(self, propKey):

        def perform(*args, **kwargs):
            print(args)
            print(kwargs)
            kwargs.setdefault('headers', {})

            target = object.__getattribute__(self, "_target")
            if propKey == 'request':
                method = args[0]
                url = args[1]
            else:
                method = propKey 
                url = args[0]

            def execAction(peer, zipkin_span=None):
                print(peer)
                print(url)
                origMethod = getattr(requests, propKey)
                newargs = []
                if propKey == 'request':
                    newargs.append(method)    
                finalUrl = peer['protocol'] + '://' + peer['address'] + ':' + str(peer['port'])
                if url:
                    finalUrl = finalUrl + url
                newargs.append(finalUrl)
                if zipkin_span:
                    zipkin.addZipkinHeaders(kwargs['headers'], zipkin_span)
                result = origMethod(*newargs, **kwargs)
                return result
            binary_annotations = {
                'http.url': url
            }
            executor = Executor(registry, policy, zipkin, target, method, binary_annotations, execAction)
            return executor.perform()

        return perform


import aws as AWS
nativeClientFetcher = {
    "dynamodb" : AWS.fetchDynamoClient,
    "kinesis" : AWS.fetchKinesisClient
}
nativeClientArgSetter = {
    "kinesis" : AWS.setupKinesisArgs
}


class NativeResourceWrapper(object):

    def __init__(self, target):
        object.__setattr__(self, "_target", target)

    def __getattribute__(self, propKey):

        def perform(*args, **kwargs):
            target = object.__getattribute__(self, "_target")

            def execAction(peer, zipkin_span=None):
                logger.info('Running %s', propKey)
                clientFetcher = nativeClientFetcher.get(peer['subClass'])
                if not clientFetcher:
                    raise Exception('Service not supported', peer['subClass'])
                client = clientFetcher(peer)
                origMethod = getattr(client, propKey)
                argsSetter = nativeClientArgSetter.get(peer['subClass'])
                if argsSetter:
                    argsSetter(peer, propKey, args, kwargs)
                result = origMethod(*args, **kwargs)
                logger.info('Running %s completed', propKey)
                return result
            executor = Executor(registry, policy, zipkin, target, propKey, None, execAction)
            return executor.perform()

        return perform


def getNativeClient(kind, name):
    nativeClient = NativeResourceWrapper([kind, name])
    return nativeClient


def instrument(method, binary_annotations):
    return zipkin.instrument(method, binary_annotations)


def randomFromList(list):
    return rand.choice(list)

def randomFromDict(dict):
    if not dict:
        return False
    key = randomFromList(dict.keys())
    return dict[key]

def setupFlask(app):
    from frameworks.b_flask import Flask
    Flask(app, zipkin, policy)


import os
from setuptools import setup

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "berlioz",
    version = "0.0.4",
    url = 'https://berlioz.cloud',
    author = "Ruben Hakopian",
    author_email = "ruben@berlioz.cloud",
    description = ("Python SDK for Berlioz Cloud"),
    license = "Apache License 2.0",
    keywords = "microservices service mesh cloud aws",
    packages=['berlioz'],
    long_description=read('README.rst'),
    platforms='any',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: Apache Software License",
    ],
)
